/*
   This file is part of Callgrind, a Valgrind skin for call graph
   profiling programs.

   Copyright (C) 2002-2004, Josef Weidendorfer (Josef.Weidendorfer@gmx.de)

   This skin is derived from and contains lot of code from Cachegrind
   Copyright (C) 2002 Nicholas Nethercote (njn25@cam.ac.uk)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#include "ct_include.h"

/*------------------------------------------------------------*/
/*--- Debug output helpers                                 ---*/
/*------------------------------------------------------------*/

static void print_indent(int s)
{
    /* max of 40 spaces */
    char sp[] = "                                        ";
    if (s>40) s=40;
    VG_(printf)(sp+40-s);
}

void SK_(print_bb)(int s, BB* bb)
{
    if (s<0) {
	s = -s;
	print_indent(s);
    }

    VG_(printf)("BB 0x%08x (Obj '%s')", bb->addr, bb->obj->name);
}

static
void print_mangled_cxt(Context* cxt, int rec_index)
{
    int i;

    if (!cxt)
      VG_(printf)("(no context)");
    else {
      VG_(printf)("%s", cxt->fn[0]->name);
      if (rec_index >0)
	VG_(printf)("'%d", rec_index +1);
      for(i=1;i<cxt->size;i++)
	VG_(printf)("'%s", cxt->fn[i]->name);
    }
}



void SK_(print_cxt)(int s, Context* cxt, int rec_index)
{
    if (s<0) {
	s = -s;
	print_indent(s);
    }

    if (cxt) {
      SK_(assert_fn_number_valid)(cxt->fn[0]->number);
      CT_ASSERT(rec_index < cxt->fn[0]->separate_recursions);

      VG_(printf)("Cxt %d (rmax %d, cmax %d) - act %d: ",
		  cxt->base_number + rec_index,
		  cxt->fn[0]->separate_recursions, cxt->size-1,
		  SK_(fn_active_array)[cxt->fn[0]->number]);
      print_mangled_cxt(cxt, rec_index);
      VG_(printf)("\n");
    }
    else
      VG_(printf)("(no context)\n");
}


void SK_(print_bbcc)(int s, BBCC* bbcc)
{
    if (s<0) {
	s = -s;
	print_indent(s);
    }

    if (!bbcc) {
	VG_(printf)("BBCC 0x0\n");
	return;
    }

    VG_(printf)("BBCC 0x%p for BB 0x%x, exe %llu, ret %llu\n",
		bbcc, bbcc->bb->addr,
		bbcc->exe_counter, bbcc->ret_counter);

    SK_(print_cxt)(-(s+8), bbcc->cxt, bbcc->rec_index);
}

void SK_(print_fcc)(int s, fCC* fcc)
{
    if (s<0) {
	s = -s;
	print_indent(s);
    }
    VG_(printf)("CC Ir A %llu, L1Miss %llu, L2Miss %llu\n",
		fcc->Ir.a, fcc->Ir.m1, fcc->Ir.m2);
    print_indent(s+3);
    VG_(printf)("Dr A %llu, L1Miss %llu, L2Miss %llu\n",
		fcc->Dr.a, fcc->Dr.m1, fcc->Dr.m2);
    print_indent(s+3);
    VG_(printf)("Dw A %llu, L1Miss %llu, L2Miss %llu\n",
		fcc->Dw.a, fcc->Dw.m1, fcc->Dw.m2);
}

void SK_(print_short_jcc)(jCC* jcc)
{
    if (jcc)
	VG_(printf)("%x => %x [%llu/%llu,%llu,%llu]",
		    jcc->from->bb->jmp_addr,
		    jcc->to->bb->addr,
		    jcc->call_counter,
		    jcc->sum.Ir.a,
		    jcc->sum.Dr.a,
		    jcc->sum.Dw.a);
    else
	VG_(printf)("[Skipped JCC]");
}

void SK_(print_jcc)(int s, jCC* jcc)
{
    if (s<0) {
	s = -s;
	print_indent(s);
    }

    if (!jcc) {
	VG_(printf)("JCC to skipped function");
	return;
    }
    VG_(printf)("JCC from ");
    SK_(print_bbcc)(s+9, jcc->from);
    print_indent(s+4);    
    VG_(printf)("to   ");
    SK_(print_bbcc)(s+9, jcc->to);
    print_indent(s+4);    
    VG_(printf)("with call count %llu\n", jcc->call_counter);
    print_indent(s+4);    
    VG_(printf)("cost ");
    SK_(print_fcc)(s+9, &(jcc->sum));
}

/* dump out the current call stack */
void SK_(print_stackentry)(int s, int sp)
{
    call_entry* ce;

    if (s<0) {
	s = -s;
	print_indent(s);
    }

    ce = SK_(get_callentry)(sp);
    VG_(printf)("[%-2d] ESP 0x%x", sp, ce->esp);
    if (ce->nonskipped)
	VG_(printf)(" NonSkipped BB 0x%x / %s",
		    ce->nonskipped->bb->addr,
		    ce->nonskipped->cxt->fn[0]->name);
    VG_(printf)("\n");
    print_indent(s+5);
    SK_(print_jcc)(5,ce->jcc);
}

/* debug output */
static void print_call_stack()
{
    int c;

    VG_(printf)("Call Stack:\n");
    for(c=0;c<SK_(get_current_sp)();c++)
      SK_(print_stackentry)(-2, c);
}

void SK_(print_bbcc_fn)(BBCC* bbcc)
{
    int i=0, opos=0;
    const UChar* obj_name;

    if (!bbcc) {
	VG_(printf)("%08x", 0);
	return;
    }

    obj_name = bbcc->cxt->fn[0]->file->obj->name;
    while(obj_name[i]) {
	if (obj_name[i]=='/') opos = i+1;
	i++;
    }

    VG_(printf)("%08x/%c  %d:", bbcc->bb->addr, 
		(bbcc->bb->sect_kind == Vg_SectText) ? 'T' :
		(bbcc->bb->sect_kind == Vg_SectData) ? 'D' :
		(bbcc->bb->sect_kind == Vg_SectBSS) ? 'B' :
		(bbcc->bb->sect_kind == Vg_SectGOT) ? 'G' :
		(bbcc->bb->sect_kind == Vg_SectPLT) ? 'P' : 'U',
		bbcc->cxt->base_number+bbcc->rec_index);
    print_mangled_cxt(bbcc->cxt, bbcc->rec_index);

    if (obj_name[0])
	VG_(printf)(" %s", obj_name+opos);

    if (VG_(strcmp)(bbcc->cxt->fn[0]->file->name, "???") !=0) {
	VG_(printf)(" %s", bbcc->cxt->fn[0]->file->name);
	if ((bbcc->cxt->fn[0] == bbcc->bb->fn) && (bbcc->bb->line>0))
	    VG_(printf)(":%d", bbcc->bb->line);
    }
}	

/* dump out an address with source info if available */
void SK_(print_addr)(Addr addr)
{
    Char fl_buf[FILENAME_LEN];
    Char fn_buf[FN_NAME_LEN];
    const UChar* obj_name;
    SegInfo* si;
    int ln, i=0, opos=0;
	
    if (addr == 0) {
	VG_(printf)("%08x", addr);
	return;
    }

    SK_(get_debug_info)(addr, fl_buf, fn_buf, &ln, &si);
    obj_name = VG_(seg_filename)(si);
    while(obj_name[i]) {
	if (obj_name[i]=='/') opos = i+1;
	i++;
    }
    
    if (VG_(strcmp)(fn_buf,"???")==0)
	VG_(printf)("%x", addr);
    else
	VG_(printf)("%x %s", addr, fn_buf);

    if (obj_name[0])
	VG_(printf)(" %s", obj_name+opos);

    if (ln>0)
    	VG_(printf)(" (%s:%u)", fl_buf,ln);
}

void SK_(print_addr_ln)(Addr addr)
{
  SK_(print_addr)(addr);
    VG_(printf)("\n");
}

static ULong bb_written = 0;

void SK_(print_bbno)(void)
{
  if (bb_written != SK_(bb_executions)) {
    bb_written = SK_(bb_executions);
    VG_(printf)("BB# %llu\n",SK_(bb_executions));
  }
}

void SK_(print_context)(void)
{
  BBCC* bbcc;

  CT_DEBUG(0,"In tid %d [%d] ",
	   SK_(get_current_tid)(),  SK_(get_current_sp)());
  bbcc =  SK_(get_current_bbcc)();
  print_mangled_cxt(SK_(get_current_cxt)(),
		    bbcc ? bbcc->rec_index : 0);
  VG_(printf)("\n");
}
