/*--------------------------------------------------------------------*/
/*--- Callgrind                                                    ---*/
/*---                                                   ct_costs.c ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of Callgrind, a Valgrind tool for call tracing.

   Copyright (C) 2002-2004, Josef Weidendorfer (Josef.Weidendorfer@gmx.de)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#include "ct_include.h" // debugging macros
#include "ct_events.h"
#include "ct_costs.h"

#define MAX_EVENTTYPE 20

static EventType eventtype[MAX_EVENTTYPE];
static Int eventtype_count = 0;

EventType* SK_(register_eventtype)(char* name)
{
  EventType* et;

  if (eventtype_count == MAX_EVENTTYPE) {
    VG_(printf)("\nMore than %d event types used!\n"
		"Increase MAX_EVENTTYPE in ct_events.c and recomile this tool!\n",
		MAX_EVENTTYPE);
    VG_(skin_panic)("Too many event types requested.");
  }

  et = &(eventtype[eventtype_count]);
  et->id = eventtype_count; 
  et->name = (UChar*) VG_(strdup)(name);
  et->description = 0;

  eventtype_count++;

  return et;
}


EventType* SK_(get_eventtype)(char* name)
{
  Int i;

  for(i=0;i<eventtype_count;i++)
    if (VG_(strcmp)(eventtype[i].name, name) == 0)
      return eventtype+i;
  return 0;
}

EventType* SK_(get_eventtype_byindex)(Int id)
{
  if ((id >= 0) && (id < eventtype_count))
    return eventtype+id;
  return 0;
}

/* Allocate space for an event set */
EventSet* SK_(get_eventset)(Int capacity)
{
  EventSet* es;

  es = (EventSet*) VG_(malloc)(sizeof(EventSet) +
			       capacity * sizeof(EventSetEntry));
  es->capacity = capacity;
  es->size = 0;

  return es;
}

/* Incorporate a event type into a set, get start offset */
Int SK_(add_eventtype)(EventSet* es, EventType* t)
{
  Int offset = es->size;
  if (es->capacity - offset < 1) return -1;

  es->size++;
  es->e[offset].type = t;
  es->e[offset].nextTop = es->size;

  return offset;
}

/* Incorporate one event set into another, get start offset */
Int SK_(add_eventset)(EventSet* dst, EventSet* src)
{
  Int offset = dst->size, i;
  if (!src || (src->size == 0)) return offset;

  if (dst->capacity - offset < src->size) return -1;
  
  for(i=0;i<src->size;i++) {
    dst->e[offset+i].type = src->e[i].type;
    dst->e[offset+i].nextTop = src->e[i].nextTop + offset;
  }
  dst->size += src->size;

  return offset;
}

/* Incorporate two event types into a set, with second < first */
Int SK_(add_dep_event2)(EventSet* es, EventType* e1, EventType* e2)
{
  Int offset = es->size;

  if (es->capacity - offset < 2) return -1;

  es->size += 2;
  es->e[offset].type = e1;
  es->e[offset].nextTop = es->size;
  es->e[offset+1].type = e2;
  es->e[offset+1].nextTop = es->size;
  
  return offset;
}

/* Incorporate 3 event types into a set, with third < second < first */
Int SK_(add_dep_event3)(EventSet* es,
			EventType* e1, EventType* e2, EventType* e3)
{
  Int offset = es->size;

  if (es->capacity - offset < 3) return -1;

  es->size += 3;
  es->e[offset].type = e1;
  es->e[offset].nextTop = es->size;
  es->e[offset+1].type = e2;
  es->e[offset+1].nextTop = es->size;
  es->e[offset+2].type = e3;
  es->e[offset+2].nextTop = es->size;
  
  return offset;
}

/* Returns number of characters written */
Int SK_(sprint_eventset)(Char* buf, EventSet* es)
{
  Int i, pos = 0;

  for(i=0; i< es->size; i++) {
    if (pos>0) buf[pos++] = ' ';
    pos += VG_(sprintf)(buf + pos, es->e[i].type->name);
  }
  buf[pos] = 0;

  return pos;
}

/* Get cost array for an event set */
ULong* SK_(get_eventset_costs)(EventSet* es)
{
  return SK_(get_costarray)(es->capacity);
}

/* Set all costs of an event set to zero */
void SK_(init_costs)(EventSet* es, ULong* costs)
{
  Int i;

  if (!costs) return;

  for(i=0;i<es->capacity;i++)
    costs[i] = 0;
}

void SK_(zero_costs)(EventSet* es, ULong* costs)
{
  Int i;

  if (!costs) return;

  for(i=0;i<es->size;i++)
    costs[i] = 0;
}
  
Bool SK_(is_zero_cost)(EventSet* es, ULong* costs)
{
  Int i = 0;

  if (!costs) return True;

  while(i<es->size) {
    if (costs[i] != 0) return False;
    i = es->e[i].nextTop;
  }
  return True;
}

Bool SK_(is_equal_cost)(EventSet* es, ULong* c1, ULong* c2)
{
  Int i = 0;

  if (!c1) return SK_(is_zero_cost)(es,c2);
  if (!c2) return SK_(is_zero_cost)(es,c1);

  while(i<es->size) {
    if (c1[i] != c2[i]) return False;
    if (c1[i] == 0)
      i = es->e[i].nextTop;
    else
      i++;
  }
  return True;
}

void SK_(copy_costs)(EventSet* es, ULong* dst, ULong* src)
{
  Int i;

  if (!src) {
    SK_(zero_costs)(es, dst);
    return;
  }
  CT_ASSERT(dst != 0);
  
  for(i=0;i<es->size;i++)
    dst[i] = src[i];
}

void SK_(copy_costs_lz)(EventSet* es, ULong** pdst, ULong* src)
{
  Int i;
  ULong* dst;

  CT_ASSERT(pdst != 0);

  if (!src) {
    SK_(zero_costs)(es, *pdst);
    return;
  }
  dst = *pdst;
  if (!dst)
    dst = *pdst = SK_(get_eventset_costs)(es);
  
  for(i=0;i<es->size;i++)
    dst[i] = src[i];
}

void SK_(add_costs)(EventSet* es, ULong* dst, ULong* src)
{
  Int i = 0;

  if (!src) return;
  CT_ASSERT(dst != 0);

  while(i<es->size) {
    if (src[i] == 0)
      i = es->e[i].nextTop;
    else {
      dst[i] += src[i];
      i++;
    }
  }
}

void SK_(add_costs_lz)(EventSet* es, ULong** pdst, ULong* src)
{
  Int i;
  ULong* dst;

  if (!src) return;
  CT_ASSERT(pdst != 0);

  dst = *pdst;
  if (!dst) {
    dst = *pdst = SK_(get_eventset_costs)(es);
    SK_(copy_costs)(es,dst,src);
    return;
  }

  i = 0;
  while(i<es->size) {
    if (src[i] == 0)
      i = es->e[i].nextTop;
    else {
      dst[i] += src[i];
      i++;
    }
  }
}

/* Adds src to dst and zeros src. Returns false if nothing changed */
Bool SK_(add_and_zero_costs)(EventSet* es, ULong* dst, ULong* src)
{
  Int i = 0, j = 0;

  if (!src) return False;

  while(i<es->size) {
    if (src[i] == 0)
      i = es->e[i].nextTop;
    else {
      dst[i] += src[i];
      src[i] = 0;
      i++;
      j++;
    }
  }

  return (j>0);
}

/* Adds src to dst and zeros src. Returns false if nothing changed */
Bool SK_(add_and_zero_costs_lz)(EventSet* es, ULong** pdst, ULong* src)
{
  Int i;
  ULong* dst;

  if (!src) return False;

  i = 0;
  while(1) {
    if (i >= es->size) return False;
    if (src[i] != 0) break;
    i = es->e[i].nextTop;
  }

  CT_ASSERT(pdst != 0);
  dst = *pdst;
  if (!dst) {
    dst = *pdst = SK_(get_eventset_costs)(es);
    SK_(copy_costs)(es,dst,src);
    SK_(zero_costs)(es,src);
    return True;
  }

  dst[i] += src[i];
  src[i] = 0;
  i++;

  while(i<es->size) {
    if (src[i] == 0)
      i = es->e[i].nextTop;
    else {
      dst[i] += src[i];
      src[i] = 0;
    }
  }

  return True;
}

/* Adds difference of new and old to dst, and set old to new.
 * Returns false if nothing changed */
Bool SK_(add_diff_costs)(EventSet* es, ULong* dst, ULong* old, ULong* new)
{
  Int i = 0, j = 0;

  while(i<es->size) {
    if (new[i] == old[i])
      i = es->e[i].nextTop;
    else {
      dst[i] += new[i] - old[i];
      old[i] = new[i];
      i++;
      j++;
    }
  }

  return (j>0);
}

/* Adds difference of new and old to dst, and set old to new.
 * Returns false if nothing changed */
Bool SK_(add_diff_costs_lz)(EventSet* es, ULong** pdst, 
			    ULong* old, ULong* new)
{
  Int i;
  ULong* dst;

  if (!old && !new) return False;
  CT_ASSERT(old && new);

  i = 0;
  while(1) {
    if (i >= es->size) return False;
    if (old[i] != new[i]) break;
    i = es->e[i].nextTop;
  }

  CT_ASSERT(pdst != 0);
  dst = *pdst;
  if (!dst) {
    dst = *pdst = SK_(get_eventset_costs)(es);
    SK_(zero_costs)(es,dst);
  }

  dst[i] += new[i] - old[i];
  old[i] = new[i];
  i++;

  while(i<es->size) {
    if (new[i] == old[i])
      i = es->e[i].nextTop;
    else {
      dst[i] += new[i] - old[i];
      old[i] = new[i];
      i++;
    }
  }

  return True;
}

/* Returns number of characters written */
Int SK_(sprint_costs)(Char* buf, EventSet* es, ULong* c)
{
  Int i = 0, pos = 0, skipped = 0;

  if (!c) return 0;

  while(i<es->size) {
    if (c[i] == 0) {
      skipped  += es->e[i].nextTop - i;
      i = es->e[i].nextTop;
    }
    else {
      while(skipped>0) {
	if (pos>0) buf[pos++] = ' ';
	buf[pos++] = '0';
	skipped--;
      }
      if (pos>0) buf[pos++] = ' ';
      pos += VG_(sprintf)(buf+pos, "%llu", c[i]);
      i++;
    }
  }

  return pos;
}

Int SK_(get_eventlist_len)(Int size)
{
  return sizeof(EventList) + (size+1) * sizeof(EventListEntry);
}

void SK_(init_eventlist)(EventList* el, Int size)
{
  el->size = size;
  el->used = 0;
  el->costcount = 0;
  /* additinal end of list marker */
  el->e[0].instr_offset = 0;
  el->e[0].cost_offset = 0;
  el->e[0].eventset = 0;
}

/* Allocate space for an event list */
EventList* SK_(get_eventlist)(Int size)
{
  EventList* el;

  el = (EventList*) VG_(malloc)(SK_(get_eventlist_len)(size));
  SK_(init_eventlist)(el, size);

  return el;
}

/* Append an eventset. Returns offset */
Int SK_(append_eventset)(EventList* el, EventSet* es, Int instr_size)
{
  Int offset = el->size;

  if (!es || (el->size - el->used <1)) return -1;

  el->used++;
  el->costcount += es->size;
  el->e[offset].eventset = es;
  el->e[offset+1].instr_offset = el->e[offset].instr_offset + instr_size;
  el->e[offset+1].cost_offset  = el->e[offset].cost_offset + es->size;
  el->e[offset+1].eventset = 0;

  return offset;
}


/* Allocate cost array for an event list */
ULong* SK_(get_eventlist_costs)(EventList* el)
{
  return SK_(get_costarray)(el->costcount);
}
